
/* This layer of software is merely concerned with representing the TA graph structures 
   Any access to update methods at this level will perform no signalling of changes
   These methods should only be used during the actual loading of an initial TA graph
 */
class CEntityInstance;
class CEntityClass;
class CRelationInstance;
class CRelationClass;
class CEntityCache;
class CLandscapeTokenStream;

class CTAInput : public Cobject {
private:
	void createRootInstance(void);
	void setupUniversalScheme(void); 
	void processSchemeTuples(CLandscapeTokenStream *tsP); 
	void processFactTuples(CLandscapeTokenStream *tsP); 
	void parseStream(void); 

protected:
	CEntityInstance	*m_rootInstanceP;				// The root instance of the diagram
	CEntityClass	*m_entityClassListP;
	CEntityClass	*m_entityBaseClassP;
	CRelationClass	*m_relationClassListP;
	CRelationClass	*m_relationBaseClassP;

public:
	CEntityCache	*m_entityCacheP;

	CEntityClass *newEntityClass(const char *idP);
	CEntityClass *getEntityClass(const char *idP);
	CEntityClass *addEntityClass(const char *idP); 
	CRelationClass *newRelationClass(const char *idP);
	CRelationClass *getRelationClass(const char *idP);
	CRelationClass *addRelationClass(const char *idP); 
	CTAInput(void); 
	CEntityClass *getEntityBaseClass(void) const;
	CRelationClass *getRelationBaseClass(void) const;
	CEntityInstance *getRootInstance(void) const;
	CEntityInstance *getCache(const char *idP) const;
	void putCache(CEntityInstance *eP) const;
	CEntityInstance *newCachedEntity(CEntityClass *ecP, const char *idP);
	void attachBaseClasses(void);
	int parseFile(const char *filenameP, int topLevel); 
	void parseStream(const char *filenameP);
	void loadTA(const char *filenameP);
};

