#define _CRT_SECURE_NO_WARNINGS

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#ifdef WIN32
#include <process.h>
#include <io.h>
#include <direct.h>

#define open _open
#define read _read
#define lseek _lseek
#define close _close

#define O_BINARY _O_BINARY
#define ctime _ctime
#else
#include <unistd.h>
#include <stdlib.h>

#define O_BINARY 0
#endif

#include "xmalloc.h"
#include "object.h"
#include "util.h"
#include "Attribute.h"
#include "AttributeCache.h"
#include "LandscapeObject.h"
#include "LandscapeObject3D.h"
#include "EntityInstance.h"
#include "RelationInstance.h"
#include "LandscapeClassObject.h"
#include "EntityClass.h"
#include "RelationClass.h"
#include "TAInput.h"
#include "boundaries.h"
#include "signature_buffer.h"
#include "signature.h"

static int g_argc    = 0;
static const char **g_argv = 0;

static CTAInput 		ta;
static CEntityInstance	*g_headP = 0;

static void visit(Clink *linkP, CEntityInstance *eP, int depth);

static void
trap()
{
}

/* For any call from f(x) to a virtual function A::g(), add as needed
 * presumed calls from f(x) to every virtual function X::g() for which
 * X is a direct subclass of A.
 */

static Csignature	signature1, signature2;

static void
handle_polymorphism(Clink *linkP, CEntityInstance *function1P, int depth)
{
	char				*mangle1P, *mangle2P;
	const char			*idP;
	CEntityInstance		*fa1P, *vtable1P, *typeinfo1P;
	CEntityInstance		*fa2P, *vtable2P, *typeinfo2P, *addrP;
	CEntityInstance		*function2P;
	CRelationInstance	*riP, *ri1P;

	if (function1P->m_visited & PolyX) {
		return;
	}

	idP = function1P->getType();
	if (!idP || strcmp(idP, "VF")) {
		// Not a virtual function
		return;
	}

#if 0
	// Check to see if it is ever called virtually

	if (direct) {
		for (riP = function1P->m_dstRelListP; ; riP = riP->m_next_dstP) {
			if (!riP) {
				// No virtual calls to this function found
				return;
			}
			idP = riP->getType();
			if (!strcmp(idP, "CVF")) {
				// There is a virtual call to this function
				break;
	}	}	}
#endif

	function1P->m_visited |= PolyX;

	mangle1P = (char *) function1P->getAttribute("mangle");
	if (!mangle1P) {
		return;
	}
	if (!signature1.signature(mangle1P)) {
		return;
	}

	// Find the variable representing the address of function1

	for (riP = function1P->m_srcRelListP; ; riP = riP->m_next_srcP) {
		if (!riP) {
			assert(riP);
			return;
		}
		idP  = riP->getType();
		if (strcmp(idP, "contain")) {
			continue;
		}
		fa1P = riP->m_dstP;
		idP  = fa1P->getType();
		if (!strcmp(idP, "FA")) {
			break;
	}	}

	// Find the vtable for this functions class
	// Note a function can be called from many vtables

	for (riP = fa1P->m_dstRelListP; ; riP = riP->m_next_dstP) {
		if (!riP) {
			assert(riP);
			return;
		}
		vtable1P = riP->m_srcP;
		idP      = vtable1P->getType();
		if (strcmp(idP, "TV")) {
			// Not a vtable
			continue;
		}
		mangle2P = (char *) vtable1P->getAttribute("mangle");
		if (!mangle2P) {
			continue;
		}
		if (!signature2.signature(mangle2P)) {
			continue;
		}

		if (signature1.same_class(&signature2)) {
			// Found correct class
			break;
	}	}

	// Find the typeinfo address of vtable1

	for (riP = vtable1P->m_srcRelListP; ; riP = riP->m_next_srcP) {
		if (!riP) {
			assert(riP);
			return;
		}
		addrP = riP->m_dstP;
		idP   = addrP->getType();
		if (!strcmp(idP, "VA")) {
			typeinfo1P = addrP->getParent();
			idP        = typeinfo1P->getType();
			if (!strcmp(idP, "TI")) {
				break;
	}	}	}

	// For each typeinfo2P which references our typeinfo1P address
	// implying that its class directly inherits from class1P

	for (riP = addrP->m_dstRelListP; riP; riP = riP->m_next_dstP) {
		typeinfo2P = riP->m_srcP;
		if (typeinfo2P == typeinfo1P) {
			continue;
		}
		idP = typeinfo2P->getType();
		if (strcmp(idP, "TI")) {
			continue;
		}
		// Find the address for typeinfo2P

		for (ri1P = typeinfo2P->m_srcRelListP; ; ri1P = ri1P->m_next_srcP) {
			if (!ri1P) {
				assert(0);
				goto skip_match;
			}
			idP = ri1P->getType();
			if (strcmp(idP, "contain")) {
				continue;
			}

			addrP = ri1P->m_dstP;
			idP   = addrP->getType();
			if (!strcmp(idP, "VA")) {
				break;
		}	}

		// Find the vtable pointing at typeinfo2P address

		for (ri1P = addrP->m_dstRelListP; ; ri1P = ri1P->m_next_dstP) {
			if (!ri1P) {
				assert(0);
				goto skip_match;
			}
			vtable2P = ri1P->m_srcP;
			idP      = vtable2P->getType();
			if (!strcmp(idP, "TV")) {
				break;
		}	}

		// Find a function in that vtable with the same signature as ours

		for (ri1P = vtable2P->m_srcRelListP; ri1P; ri1P = ri1P->m_next_srcP) {
			fa2P = ri1P->m_dstP;
			idP  = fa2P->getType();
			if (strcmp(idP, "FA")) {
				continue;
			}
			function2P = fa2P->getParent();
			idP        = function2P->getType();
			if (strcmp(idP, "VF")) {
				continue;
			}
			mangle2P = (char *) function2P->getAttribute("mangle");
			if (!mangle2P) {
				continue;
			}
			if (!signature2.signature(mangle2P)) {
				continue;
			}
			if (signature1.same_function_name(&signature2)) {
				if (!(function2P->m_visited & CalledX)) {
					visit(linkP, function2P, depth+1);
				}
		}	}
skip_match:
		continue;
	}
}

static int
same(CEntityInstance *e1P, CEntityInstance *e2P)
{
	CEntityClass	*ec1P, *ec2P;

	ec1P = e1P->getEntityClass();
	ec2P = e2P->getEntityClass();
	if (ec1P != ec2P) {
		return 0;
	}
	if (strcmp(e1P->getLabel(), e2P->getLabel())) {
		return 0;
	}
	return 1;
}

/* Reduce all references to a function (in a header file) seen in amy
 * source files to a single reference. Don't want to treat a function
 * in a header as dead code unless there really are no links to it
 * from any source file
 */

static void
consolidate(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP, **ePP, *e1P;
	const char			*labelP;
	CRelationInstance	*riP, *nextP;
	const char			*fullpathP;
	char				*P;
	int					diff;

	// Sort by source file name and then by lineno
	for (ecP = ta.getEntityClass(0); ecP; ecP = ecP->m_nextP) {
		labelP = ecP->getId();
		if (!labelP) {
			continue;
		}
		if (strcmp(labelP, "GF") && 
            strcmp(labelP, "LF") &&
			strcmp(labelP, "VF")) {
			continue;
		}

		for (eP = ecP->m_entityInstancesP; eP; eP = eP->m_nextInClassP) {
			fullpathP = eP->m_fullpathP;
			if (!fullpathP || !strcmp(fullpathP, ".")) {
				P = 0;
				eP->buildPath(&P, 0);
				P = (char *) Xmalloc(P - ((char *) 0) + 1);
				eP->m_fullpathP = P;
				eP->buildPath(&P, 1);
			}
			fullpathP = eP->m_fullpathP;

			for (ePP = &g_headP; (e1P = *ePP); ePP = &(e1P->m_nextInOrderP)) {
				diff = strcmp(fullpathP, e1P->m_fullpathP);
				if (diff < 0) {
					goto found;
				}
				if (diff) {
					continue;
				}
				if (eP->m_lineno < e1P->m_lineno) {
					goto found;
				}
				if (eP->m_lineno == e1P->m_lineno && same(eP, e1P)) {
					// if source is compiled more than once
					// The same function might be named multiple times
					//printf("Consolidating %d:[%s]%s/[%s]%s in %s\n", eP->m_lineno, eP->getId(), eP->getLabel(), e1P->getId(), e1P->getLabel(), eP->m_fullpathP);
					for (riP = eP->m_srcRelListP; riP; riP = nextP) {
						nextP = riP->m_next_srcP;
						// Move edge to be sourced by earlier node
						riP->m_next_srcP   = e1P->m_srcRelListP;
						e1P->m_srcRelListP = riP;
						riP->m_srcP        = e1P;
					}
					for (riP = eP->m_dstRelListP; riP; riP = nextP) {
						nextP = riP->m_next_dstP;
						// Moce edge to target earlier node
						riP->m_next_dstP   = e1P->m_dstRelListP;
						e1P->m_dstRelListP = riP;
						riP->m_dstP        = e1P;
					}
					eP->m_srcRelListP = 0;
					eP->m_dstRelListP = 0;
					eP->m_visited     = SkippedX;
					goto skip;
			}	}	
found:		eP->m_nextInOrderP = *ePP;
			*ePP               = eP;
skip: 		continue;
	}	}	
}

static void
dumppath(Clink *linkP, int depth)
{
	CEntityInstance	*eP = linkP->m_entityP;
	int	i;

	if (linkP->m_parentP) {
		dumppath(linkP->m_parentP, depth-1);
	}
	for (int i = 0; i < depth; ++i) {
		fputs("  ", stdout);
	}
	fprintf(stdout, " %s", eP->getLabel());
	if (eP->m_lineno != 0x7fffffff) {
		fprintf(stdout, " %d", eP->m_lineno);
	}
	if (eP->m_fullpathP) {
		fprintf(stdout, ":%s", eP->m_fullpathP);
	}
	fputc('\n', stdout);
}

static void
visit(Clink *linkP, CEntityInstance *eP, int depth)
{
	Clink link(linkP, eP);

	CRelationInstance	*riP;
	const char			*mangleP;
	const char			*labelP, *typeP;
	CEntityInstance		*dstP;
	int					i;

    if (eP->m_visited) {
      return;
    }
	mangleP = eP->getAttribute("mangle");
	if (mangleP) {
		if (!strncmp(mangleP, "_ZT", 3)) {
			// Don't visit under thunks/vtables/etc
			return;
		}
	}
	eP->m_visited |= CalledX;
	typeP         = eP->getType();

	labelP = eP->getLabel();
	for (i = 2; i < g_argc; ++i) {
		if (!strcmp(labelP, g_argv[i])) {
			dumppath(&link, depth);
		}
	}

	switch(typeP[0]) {
	case 'V':
	case 'F':
		if (typeP[1] == 'A') {
			dstP = eP->getParent();
			if (dstP && !dstP->m_visited) {
				visit(&link, dstP, depth+1);
	}	}	}

	for (riP = eP->m_srcRelListP; riP; riP = riP->m_next_srcP) {
		labelP = riP->getType();
		dstP   = riP->m_dstP;
		switch (labelP[0]) {
		case 'C':
			if (!strcmp(labelP, "contain")) {
				continue;
			}
		default:
			// Even variables might reference the address of functions
			// For example: int fred() {}; int (*fredP)() = fred;
			break;
		} 
		if (!dstP->m_visited) {
			visit(&link, dstP, depth+1);
	}	}

	handle_polymorphism(&link, eP, depth);
}

static void
depth_search(void)
{
	CEntityInstance		*eP;
	const char			*labelP;

	// Find the functions and variables not visited
	for (eP = g_headP; eP; eP = eP->m_nextInOrderP) {
		labelP = eP->getLabel();

		/* Some functions called during initialistation */
		if (*labelP == '"') {
			++labelP;
		}
		if (!strncmp(labelP, "global ", 7) ||
			!strncmp(labelP, "__static_", 9)) {
			goto descend;
		}
		if (strcmp(labelP, "main") && strcmp(labelP, "wmain")) {
			continue;
		}
		labelP = eP->getType();
		if (!labelP) {
			continue;
		}
		if (strcmp(labelP, "GF")) {
			continue;
		}
descend:
		if (!eP->m_visited) {
			visit(0, eP, 0);
}	}	}

#ifndef SSIZE_MAX
#define SSIZE_MAX 0x7FFFFFFF
#endif

const char	*g_filenameP = "";
static char	*g_sourceP   = 0;
static int	g_source_lth = 0;

static int
test_template(CEntityInstance *eP)
{
	CEntityInstance		*srcP, *dstP;
	CRelationInstance	*riP, *ri1P;
	const char			*idP;
	int					ret;

	ret = 0;
	for (riP = eP->m_dstRelListP; riP; riP = riP->m_next_dstP) {
		srcP = riP->m_srcP;
		idP  = srcP->getType();
		if (strcmp(idP, "T")) {
			continue;
		}
		ret = 1;
		for (ri1P = srcP->m_srcRelListP; ri1P; ri1P = ri1P->m_next_srcP) {
			dstP = ri1P->m_dstP;
			if (!dstP->m_visited) {
				ret = 2;
				break;
	}	}	}
	return ret;
}

static int
is_top(CEntityInstance *eP)
{
	CEntityInstance		*srcP, *dstP;
	CRelationInstance	*riP;
	const char			*idP, *typeP;

	for (riP = eP->m_dstRelListP; riP; riP = riP->m_next_dstP) {
		idP  = riP->getType();
		if (!strcmp(idP, "contain")) {
			continue;
		}
		srcP = riP->m_srcP;
		if (!srcP->m_visited) {	
			// Something not visited calls me
			return 0;
	}	}
	for (riP = eP->m_srcRelListP; riP; riP = riP->m_next_srcP) {
		idP = riP->getType();
		if (strcmp(idP, "contain")) {
			continue;
		}
		dstP = riP->m_dstP;
		typeP = dstP->getType();
		switch(typeP[0]) {
		case 'V':
		case 'F':
			if (typeP[1] == 'A') {
				return(is_top(dstP));
	}	}	}
	return 1;
}

static int
overrides(CEntityInstance *typeinfo1P)
{
	char				*mangle2P;
	const char			*idP;
	CEntityInstance		*fa2P, *vtable2P, *typeinfo2P, *addrP;
	CEntityInstance		*function2P;
	CRelationInstance	*riP, *ri1P;

	// Find a typeinfo sturct that typeinfo1P points at

	for (riP = typeinfo1P->m_srcRelListP; riP; riP = riP->m_next_srcP) {
		typeinfo2P = riP->m_dstP;
		if (typeinfo2P == typeinfo1P) {
			continue;
		}
		idP = typeinfo2P->getType();
		if (strcmp(idP, "TI")) {
			continue;
		}
		// Find the address for typeinfo2P

		for (ri1P = typeinfo2P->m_srcRelListP; ; ri1P = ri1P->m_next_srcP) {
			if (!ri1P) {
				assert(0);
				goto skip_match;
			}
			idP = ri1P->getType();
			if (strcmp(idP, "contain")) {
				continue;
			}

			addrP = ri1P->m_dstP;
			idP   = addrP->getType();
			if (!strcmp(idP, "VA")) {
				break;
		}	}

		// Find the vtable pointing at typeinfo2P address

		for (ri1P = addrP->m_dstRelListP; ; ri1P = ri1P->m_next_dstP) {
			if (!ri1P) {
				assert(0);
				goto skip_match;
			}
			vtable2P = ri1P->m_srcP;
			idP      = vtable2P->getType();
			if (!strcmp(idP, "TV")) {
				break;
		}	}

		// Find a function in that vtable with the same signature as ours

		for (ri1P = vtable2P->m_srcRelListP; ri1P; ri1P = ri1P->m_next_srcP) {
			fa2P = ri1P->m_dstP;
			idP  = fa2P->getType();
			if (strcmp(idP, "FA")) {
				continue;
			}
			function2P = fa2P->getParent();
			idP        = function2P->getType();
			if (strcmp(idP, "VF")) {
				continue;
			}
			mangle2P = (char *) function2P->getAttribute("mangle");
			if (!mangle2P) {
				continue;
			}
			if (!signature2.signature(mangle2P)) {
				continue;
			}
			if (signature1.same_function_name(&signature2)) {
				return 1;
		}	}
		if (overrides(typeinfo2P)) {
			return 1;
		}
skip_match:
		continue;
	}
	return 0;
}

#ifdef TODO
static void
overrides_virtual_function(CEntityInstance *function1P)
{
	char				*mangle2P;
	const char			*idP;
	CEntityInstance		*fa1P, *vtable1P, *typeinfo1P;
	CEntityInstance		*addrP;
	CRelationInstance	*riP;

	idP = function1P->getType();
	if (!idP || strcmp(idP, "VF")) {
		// Not a virtual function
		return;
	}

	for (riP = function1P->m_srcRelListP; ; riP = riP->m_next_srcP) {
		if (!riP) {
			assert(riP);
			return;
		}
		idP  = riP->getType();
		if (strcmp(idP, "contain")) {
			continue;
		}
		fa1P = riP->m_dstP;
		idP  = fa1P->getType();
		if (!strcmp(idP, "FA")) {
			break;
	}	}

	// Find the vtable for this functions class
	// Note a function can be called from many vtables

	for (riP = fa1P->m_dstRelListP; ; riP = riP->m_next_dstP) {
		if (!riP) {
			assert(riP);
			return;
		}
		vtable1P = riP->m_srcP;
		idP      = vtable1P->getType();
		if (strcmp(idP, "TV")) {
			// Not a vtable
			continue;
		}
		mangle2P = (char *) vtable1P->getAttribute("mangle");
		if (!mangle2P) {
			continue;
		}
		if (!signature2.signature(mangle2P)) {
			continue;
		}

		if (signature1.same_class(&signature2)) {
			// Found correct class
			break;
	}	}


	// Find the typeinfo address of vtable1

	for (riP = vtable1P->m_srcRelListP; ; riP = riP->m_next_srcP) {
		if (!riP) {
			assert(riP);
			return;
		}
		addrP = riP->m_dstP;
		idP   = addrP->getType();
		if (!strcmp(idP, "VA")) {
			typeinfo1P = addrP->getParent();
			idP        = typeinfo1P->getType();
			if (!strcmp(idP, "TI")) {
				break;
	}	}	}

	// For each typeinfo2P referenced directly or indirectly from typeinfo1P
	// implying that its class is a superclass see if it also declares this
	// function

	if (overrides(typeinfo1P)) {
		visit(function1P, 0);
		function1P->m_visited &= ~VisitedX;
		function1P->m_visited |= OverridesX;
}	}
#endif

static void
find_path(const char *filenameP)
{
	ta.loadTA(filenameP);
	consolidate();
	depth_search();
}

int 
main(int argc, const char *argv[])
{
	const char *filenameP = 0;
	int			i;
	const char	*argP;
	char		cwd[2024];

	g_argc = argc;
	g_argv = argv;

	filenameP = argv[1];

	printf("%s " VERSION " " __DATE__ " " __TIME__ "\n", argv[0]);
	find_path(filenameP);

	printf("\n(Press <CR> to continue\n");
	getchar();
	return 0;
}

